"""
Configuration settings for RF-DETR export toolkit
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List


@dataclass
class ExportConfig:
    """Configuration for model export"""

    default_imgsz: int = 560
    default_precision: str = "fp16"
    onnx_opset: int = 17
    simplify_onnx: bool = True


@dataclass
class MetadataConfig:
    """Configuration for metadata generation"""

    schema_version: int = 1
    model_type: str = "DetectionModel"
    version: int = 1
    default_icon: str = "Pending"

    # Preprocessing defaults
    maintain_aspect_ratio: bool = False
    symmetric_padding: bool = False
    padding_fill_color: list[int] = None
    normalize_mean: list[float] = None
    normalize_std: list[float] = None

    # Detection thresholds
    default_threshold: float = 0.5
    low_threshold: float = 0.25
    medium_threshold: float = 0.5
    high_threshold: float = 0.75

    def __post_init__(self):
        if self.padding_fill_color is None:
            self.padding_fill_color = [114, 114, 114]
        if self.normalize_mean is None:
            self.normalize_mean = [0.485, 0.456, 0.406]
        if self.normalize_std is None:
            self.normalize_std = [0.229, 0.224, 0.225]


# Global configuration instances
EXPORT_CONFIG = ExportConfig()
METADATA_CONFIG = MetadataConfig()



"""
Utility functions for export toolkit
"""
from __future__ import annotations

import logging
import os
import random
import zipfile
from pathlib import Path
from typing import List


def setup_logging(verbose: bool = False):
    """Setup logging configuration"""
    level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        level=level,
        format="%(asctime)s - %(levelname)s - %(message)s",
        handlers=[logging.StreamHandler()],
    )


def validate_inputs(args):
    """Validate command line arguments"""
    # Validate model path (if not using default)
    if not Path(args.model).exists():
        raise ValueError(f"Model file not found: {args.model}")

    # Validate class names
    if not args.classes or not args.classes.strip():
        raise ValueError("Class names are required (--classes)")

    class_names = [name.strip().upper() for name in args.classes.split(",")]
    if not all(name for name in class_names):
        raise ValueError("All class names must be non-empty")

    # Validate class indices if provided
    if args.anonymize_classes:
        indices = [int(x.strip()) for x in args.anonymize_classes.split(",")]
        if any(i >= len(class_names) or i < 0 for i in indices):
            raise ValueError("Anonymize class indices out of range")

    if args.georeference_classes:
        indices = [
            int(x.strip()) for x in args.georeference_classes.split(",")
        ]
        if any(i >= len(class_names) or i < 0 for i in indices):
            raise ValueError("Georeference class indices out of range")
    if args.arch:
        if args.arch.lower() not in ["rfdetr"]:
            raise ValueError(
                f"""
                ╔══════════════════════════════════════════════════╗
                ║  🚀  We currently only support RFDETR            ║
                ║                                                  ║
                ║  For assistance with exporting other models,     ║
                ║  please email: support@isarsoft.com              ║
                ╚══════════════════════════════════════════════════╝
                """
            )

    if args.arch.lower() in ["rfdetr"]:
        if args.imgsz % 56 != 0:
            raise ValueError(
                f"RF-DETR expects the input resolution to be a multiple of 56. Instead found {args.imgsz}."
            )


def generate_random_color() -> str:
    """Generate a random hex color"""
    return f"#{random.randint(0, 255):02x}{random.randint(0, 255):02x}{random.randint(0, 255):02x}".upper()


def generate_class_colors(num_classes: int, seed: int = None) -> list[str]:
    """Generate random colors for classes"""
    if seed is not None:
        random.seed(seed)

    colors = []
    for _ in range(num_classes):
        colors.append(generate_random_color())

    return colors


def create_zip_package(output_dir: str) -> str:
    """Create zip package containing exported files"""
    output_path = Path(output_dir)
    zip_path = output_path.parent / f"{output_path.name}.zip"

    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
        for file_path in output_path.glob("*"):
            if file_path.is_file():
                zipf.write(file_path, file_path.name)

    return str(zip_path)


def ensure_directory(directory: str) -> Path:
    """Ensure directory exists, create if needed"""
    path = Path(directory)
    path.mkdir(parents=True, exist_ok=True)
    return path



#!/usr/bin/env python3
"""
Model Export Toolkit
Main entry point for exporting models to ONNX with metadata
"""
from __future__ import annotations

import argparse
import sys
from pathlib import Path

from .exporter import RFDETRExporter
from .utils import setup_logging
from .utils import validate_inputs


def create_parser():
    """Create command-line argument parser"""
    parser = argparse.ArgumentParser(
        description="Export RF-DETR models to ONNX with metadata generation",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        # TODO: Fix the documentation for --arch flag
        epilog="""
Examples:
  # Basic export with class names
  python main.py --model rf-detr-base.pt --output ./exported --classes person,car,bike --model-name rfdetr

  
  
  # Export with custom metadata
  python main.py --model model.pt --output ./exported \\
    --classes person,vehicle,animal \\
    --description "Multi-class detector" \\
    --company "MyCompany" \\
    --anonymize-classes 0 \\
    --georeference-classes 1,2 \\
    --model-name yolov11

  # Export with optimization options
  python main.py --model model.pt --output ./exported \\
    --classes person \\
    --half  --imgsz 640 \\
    --model-name yolov11
        """,
    )

    # Required arguments
    parser.add_argument(
        "--model",
        "-m",
        required=True,
        help="Path to model file (.pt)",
    )

    parser.add_argument(
        "--output",
        "-o",
        required=True,
        help="Output directory for exported files",
    )

    parser.add_argument(
        "--classes",
        "-c",
        required=True,
        help='Comma-separated list of class names (e.g., "person,car,bike")',
    )

    parser.add_argument(
        "--arch",
        "-a",
        default="rfdetr",
        help="Model Architeture to Export, should be in [rfdetr]",
    )

    parser.add_argument(
        "--model-name",
        "-n",
        default="Generated",
        help="Model Name for the Frontend",
    )

    # Metadata options
    parser.add_argument(
        "--description",
        "-d",
        default="Object detection model",
        help="Model description for metadata",
    )

    parser.add_argument(
        "--company", default="Generated", help="Company name for metadata"
    )

    parser.add_argument(
        "--anonymize-classes",
        help='Comma-separated class indices to anonymize (e.g., "0,1")',
    )

    parser.add_argument(
        "--georeference-classes",
        help='Comma-separated class indices to georeference (e.g., "2,3")',
    )

    # Export options
    parser.add_argument(
        "--imgsz",
        type=int,
        default=560,
        help="Input image size for export (default: 560)",
    )

    parser.add_argument(
        "--thumbnail",
        type=str,
        default=None,
        help="Location of thumbnail for Frontend",
    )

    # Output options
    parser.add_argument(
        "--no-zip", action="store_true", help="Skip creating zip package"
    )

    parser.add_argument(
        "--verbose", "-v", action="store_true", help="Enable verbose logging"
    )

    return parser


def parse_class_indices(indices_str):
    """Parse comma-separated class indices"""
    if not indices_str:
        return []
    try:
        return [int(x.strip()) for x in indices_str.split(",")]
    except ValueError:
        raise ValueError(f"Invalid class indices: {indices_str}")


def main():
    """Main entry point"""
    parser = create_parser()
    args = parser.parse_args()

    # Setup logging
    setup_logging(verbose=args.verbose)

    try:
        # Validate inputs
        validate_inputs(args)

        # Parse class names and indices
        class_names = [name.strip() for name in args.classes.split(",")]
        anonymize_classes = parse_class_indices(args.anonymize_classes)
        georeference_classes = parse_class_indices(args.georeference_classes)

        print(f"🚀 Starting model export...")
        print(f"📋 Classes: {class_names}")
        print(f"📁 Output: {args.output}")

        if "rfdetr" in args.arch:
            exporter = RFDETRExporter(
                model_path=args.model,
                output_dir=args.output,
                arch=args.arch,
                model_name=args.model_name,
                class_names=class_names,
                description=args.description,
                company_name=args.company,
                anonymize_classes=anonymize_classes,
                georeference_classes=georeference_classes,
                thumbnail=args.thumbnail,
            )

            result = exporter.export(
                imgsz=args.imgsz,
                create_zip=not args.no_zip,
            )

        print(f"\n🎉 Export completed successfully!")
        print(f"📁 Output: {result}")

    except Exception as e:
        print(f"\n❌ Export failed: {e}")
        if args.verbose:
            import traceback

            traceback.print_exc()
        sys.exit(1)


if __name__ == "__main__":
    main()



"""
Metadata generation for RF-DETR models
"""
from __future__ import annotations

import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List

from .config import METADATA_CONFIG
from .utils import generate_class_colors


logger = logging.getLogger(__name__)


class MetadataGenerator:
    """Generate JSON metadata for exported RF-DETR models"""

    def __init__(self, output_dir: str):
        self.output_dir = Path(output_dir)
        self.config = METADATA_CONFIG

    def create_metadata_json(
        self,
        model_name: str,
        class_names: list[str],
        model_metadata: dict[str, Any],
        description: str = "RF-DETR object detection model",
        company_name: str = "Generated",
        anonymize_classes: list[int] = None,
        georeference_classes: list[int] = None,
        color_seed: int = None,
    ) -> str:
        """Create metadata JSON file with class names and random colors"""

        logger.info(f"Generating metadata for {len(class_names)} classes...")

        # Generate random colors for classes
        class_colors = generate_class_colors(len(class_names), seed=color_seed)

        # Build metadata structure
        metadata = self._build_metadata(
            model_name=model_name,
            class_names=class_names,
            class_colors=class_colors,
            model_metadata=model_metadata,
            description=description,
            company_name=company_name,
            anonymize_classes=anonymize_classes or [],
            georeference_classes=georeference_classes or [],
        )

        # Save JSON file
        json_path = self.output_dir / "config.json"
        with open(json_path, "w") as f:
            json.dump(metadata, f, ensure_ascii=False, indent=4)

        logger.info(f"✓ Generated metadata JSON: {json_path}")
        return str(json_path)

    def _build_metadata(
        self,
        model_name: str,
        class_names: list[str],
        class_colors: list[str],
        model_metadata: dict[str, Any],
        description: str,
        company_name: str,
        anonymize_classes: list[int],
        georeference_classes: list[int],
    ) -> dict[str, Any]:
        """Build the complete metadata structure"""

        timestamp = datetime.now().isoformat() + "Z"
        num_classes = len(class_names)

        # Generate model name based on classes
        if model_name is None:
            if num_classes == 1:
                model_name = class_names[0].upper()
            else:
                model_name = "MULTI_CLASS"

        out = {
            "schema_version": self.config.schema_version,
            "type": self.config.model_type,
            "version": self.config.version,
            "name": model_name,
            "description": description,
            "metadata": {
                "creator": company_name,
                "release_date": timestamp,
                "architecture": model_metadata.get("architecture", "RF-DETR"),
                "precision": model_metadata.get("precision", "fp16"),
            },
            "classes": [
                {
                    "name": class_names[i],
                    "color": class_colors[i],
                    "icon": self.config.default_icon,
                }
                for i in range(num_classes)
            ],
            "default_detection_classes": list(range(num_classes)),
            "default_tracking_groups": [],
            "default_anonymize_classes": anonymize_classes,
            "default_georeference_classes": georeference_classes,
            "default_belonging_pairs": [],
            "default_group_detection_classes": [],
        }
        if "yolo" in model_name.lower():
            out.update(
                {
                    "maintain_aspect_ratio": True,
                    "symmetric_padding": True,
                    "padding_fill_color": self.config.padding_fill_color,
                    "normalize_mean": [0.0, 0.0, 0.0],
                    "normalize_std": [1, 1, 1],
                    "detection_threshold": {
                        "default": 0.25,
                        "low": self.config.low_threshold,
                        "medium": self.config.medium_threshold,
                        "high": self.config.high_threshold,
                    },
                }
            )
        else:
            out.update(
                {
                    "maintain_aspect_ratio": self.config.maintain_aspect_ratio,
                    "symmetric_padding": self.config.symmetric_padding,
                    "padding_fill_color": self.config.padding_fill_color,
                    "normalize_mean": self.config.normalize_mean,
                    "normalize_std": self.config.normalize_std,
                    "detection_threshold": {
                        "default": self.config.default_threshold,
                        "low": self.config.low_threshold,
                        "medium": self.config.medium_threshold,
                        "high": self.config.high_threshold,
                    },
                }
            )
        return out



"""
RF-DETR model exporter with ONNX conversion and metadata generation
"""
from __future__ import annotations

import logging
import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .config import EXPORT_CONFIG
from .metadata import MetadataGenerator
from .utils import create_zip_package
from .utils import ensure_directory


logger = logging.getLogger(__name__)


# TODO: Remove the default option.
class RFDETRExporter:
    """RF-DETR model exporter with metadata generation"""

    def __init__(
        self,
        model_path: str,
        output_dir: str,
        arch: str,
        model_name: str,
        class_names: list[str],
        description: str = "RF-DETR object detection model",
        company_name: str = "Generated",
        anonymize_classes: list[int] | None = None,
        georeference_classes: list[int] | None = None,
        thumbnail: str | None = None,
    ):
        self.model_path = model_path
        self.output_dir = ensure_directory(output_dir)
        self.model_name = model_name
        self.arch = arch
        self.class_names = class_names
        self.description = description
        self.company_name = company_name
        self.anonymize_classes = anonymize_classes or []
        self.georeference_classes = georeference_classes or []
        self.thumbnail = (
            thumbnail
            if thumbnail
            else os.path.join(
                os.path.dirname(os.path.abspath(__file__)), "thumbnail.svg"
            )
        )

        self.model = None
        self.onnx_path = None

        logger.info(
            f"Initialized RF-DETR exporter for {len(class_names)} classes"
        )

    def load_model(self) -> None:
        """Load RF-DETR model"""
        try:
            # Try to import RF-DETR
            import rfdetr
            from .patch import RFDETRPatcher

            RFDETRPatcher(rfdetr)()
            # Load model
            logger.info(f"Loading RF-DETR model from {self.model_path}")
            self.model = rfdetr.RFDETRBase(pretrain_weights=self.model_path)
            logger.info("✓ RF-DETR model loaded successfully")

        except Exception as e:
            raise RuntimeError(f"Failed to load RF-DETR model: {e}")

    def export_to_onnx(
        self,
        imgsz: int = 640,
        half: bool = True,
    ) -> str:
        """Export RF-DETR model to ONNX format"""
        try:
            logger.info(
                f"Exporting RF-DETR to ONNX (imgsz={imgsz}, half={half})..."
            )

            try:
                # Export model (RF-DETR's export method)

                self.model.export(output_dir=self.output_dir, simplify=True)

                # Find exported ONNX file
                onnx_files = list(self.output_dir.glob("*.onnx"))
                if not onnx_files:
                    raise RuntimeError(
                        "ONNX export completed but no .onnx file found"
                    )

                # Use most recent if multiple files
                if len(onnx_files) > 1:
                    onnx_files.sort(
                        key=lambda x: x.stat().st_mtime, reverse=True
                    )
                    logger.warning(
                        f"Multiple ONNX files found, using: {onnx_files[0].name}"
                    )

                # Rename to standard name
                target_path = self.output_dir / "model.onnx"
                if onnx_files[0] != target_path:
                    shutil.move(str(onnx_files[0]), str(target_path))

                self.onnx_path = str(target_path)

            except Exception as e:
                raise RuntimeError(f"Failed to export RF-DETR to ONNX: {e}")

            logger.info(f"✓ RF-DETR exported to ONNX: {self.onnx_path}")
            return self.onnx_path

        except Exception as e:
            raise RuntimeError(f"Failed to export RF-DETR to ONNX: {e}")

    def get_model_metadata(self) -> dict[str, Any]:
        """Extract metadata from RF-DETR model"""
        metadata = {
            "architecture": "RF-DETR",
            "framework": "PyTorch",
            "num_classes": len(self.class_names),
            "precision": (
                "fp16" if EXPORT_CONFIG.default_precision == "fp16" else "fp32"
            ),
            "input_size": EXPORT_CONFIG.default_imgsz,
        }

        # Try to extract additional metadata from model
        try:
            if hasattr(self.model, "model"):
                if hasattr(self.model.model, "nc"):
                    detected_classes = self.model.model.nc
                    if detected_classes != len(self.class_names):
                        logger.warning(
                            f"Model has {detected_classes} classes but {len(self.class_names)} class names provided"
                        )
        except Exception as e:
            logger.debug(f"Could not extract detailed model metadata: {e}")

        return metadata

    def generate_metadata(self) -> str:
        """Generate metadata JSON file"""
        model_metadata = self.get_model_metadata()

        generator = MetadataGenerator(str(self.output_dir))
        json_path = generator.create_metadata_json(
            model_name=self.model_name,
            class_names=self.class_names,
            model_metadata=model_metadata,
            description=self.description,
            company_name=self.company_name,
            anonymize_classes=self.anonymize_classes,
            georeference_classes=self.georeference_classes,
            color_seed=42,  # Fixed seed for reproducible colors
        )

        return json_path

    def export(
        self, imgsz: int = 640, half: bool = True, create_zip: bool = True
    ) -> str:
        """Complete export workflow"""

        # Load model
        self.load_model()

        # Export to ONNX
        onnx_path = self.export_to_onnx(imgsz=imgsz, half=half)

        onnx_path = _export_slim(onnx_path=onnx_path)

        # Generate metadata
        json_path = self.generate_metadata()

        thumbnail_path = _copy_thumbnail(self.thumbnail, self.output_dir)

        logger.info(f"✓ Export completed:")
        logger.info(f"  - ONNX model: {onnx_path}")
        logger.info(f"  - Metadata: {json_path}")
        logger.info(f"  - Thumbnail: {thumbnail_path}")

        # Create zip package if requested
        if create_zip:
            zip_path = create_zip_package(str(self.output_dir))
            logger.info(f"  - Package: {zip_path}")
            return zip_path

        return str(self.output_dir)


def _export_slim(onnx_path: str):
    try:
        result = subprocess.run(
            ["onnxslim", onnx_path, onnx_path],
            check=True,  # Raises CalledProcessError on non-zero exit
            capture_output=True,  # Capture stdout/stderr
            text=True,  # Decode bytes to str
        )
    except Exception as e:
        raise RuntimeError(f"Failed to slim ONNX: {e}")

    return onnx_path


def _copy_thumbnail(thumbnail: str, output_path: str):
    if ".svg" in os.path.basename(thumbnail):
        output_path = os.path.join(output_path, "thumbnail.svg")
    elif ".png" in os.path.basename(thumbnail):
        output_path = os.path.join(output_path, "thumbnail.png")
    elif ".jpg" in os.path.basename(thumbnail):
        output_path = os.path.join(output_path, "thumbnail.jpg")
    else:
        raise RuntimeError(
            f"Unsupported thumbnail file format, use jpg, svg or png"
        )

    shutil.copyfile(thumbnail, output_path)



from __future__ import annotations

import importlib
import os
import pkgutil

import onnx
import onnxsim
import torch
from rfdetr.deploy._onnx.optimizer import OnnxOptimizer
from rfdetr.models.transformer import gen_encoder_output_proposals
from rfdetr.util import box_ops
from torch import nn
from transformers.utils import torch_int


def import_all_submodules(package_name):
    """Recursively import all submodules of a package"""
    package = importlib.import_module(package_name)

    for loader, name, is_pkg in pkgutil.walk_packages(
        package.__path__, package.__name__ + "."
    ):
        try:
            importlib.import_module(name)
        except Exception as e:
            print(f"Failed to import {name}: {e}")


class RFDETRPatcher:
    def __init__(self, module):
        self.module = module
        import_all_submodules("rfdetr")

    def __call__(self):
        # Patch dinov2 backbone export function to exclude position embedding from the export
        self.module.models.backbone.dinov2.DinoV2.export = export_backbone

        # patch main export function to enable dynamic batches and change output names
        self.module.main.export = export_main

        # patch model output to enable yolo like output
        self.module.models.lwdetr.LWDETR.forward_export = forward_export

        # patch window attention antialias
        self.module.models.backbone.dinov2_with_windowed_attn.WindowedDinov2WithRegistersEmbeddings.interpolate_pos_encoding = (
            interpolate_pos_encoding
        )

        # patch static shapes in transformer encoder
        self.module.models.transformer.Transformer.forward = forward

        # patch model export name
        self.module.deploy.export.export_onnx = export_onnx_main

        # patch onnx simplify name
        self.module.deploy.export.onnx_simplify = onnx_simplify_main


# RFDETR functions


def export_backbone(self):
    self._export = True


def export_main(
    self,
    output_dir="output",
    infer_dir=None,
    simplify=False,
    backbone_only=False,
    opset_version=17,
    verbose=True,
    force=False,
    shape=None,
    batch_size=1,
    **kwargs,
):
    """Export the trained model to ONNX format"""
    print(f"Exporting model to ONNX format")
    try:
        from rfdetr.deploy.export import make_infer_image
    except ImportError:
        print(
            "It seems some dependencies for ONNX export are missing. Please run `pip install rfdetr[onnxexport]` and try again."
        )
        raise

    device = self.device
    model = deepcopy(self.model.to("cpu"))
    model.to(device)

    os.makedirs(output_dir, exist_ok=True)
    output_dir = Path(output_dir)
    if shape is None:
        shape = (self.resolution, self.resolution)
    else:
        if shape[0] % 14 != 0 or shape[1] % 14 != 0:
            raise ValueError("Shape must be divisible by 14")

    input_tensors = make_infer_image(infer_dir, shape, batch_size, device).to(
        device
    )
    input_names = ["images"]
    output_names = (
        ["features"]
        if backbone_only
        else ["num_dets", "det_boxes", "det_scores", "det_classes"]
    )
    dynamic_axes = {
        "images": {0: "batch"},
    }
    self.model.eval()
    with torch.no_grad():
        if backbone_only:
            features = model(input_tensors)
            dynamic_axes.update(
                {
                    "features": {0: "batch"},
                }
            )
            print(f"PyTorch inference output shape: {features.shape}")
        else:
            outputs = model(input_tensors)
            dynamic_axes.update(
                {
                    "num_dets": {0: "batch"},
                    "det_boxes": {0: "batch"},
                    "det_scores": {0: "batch"},
                    "det_classes": {0: "batch"},
                }
            )
            dets = outputs["pred_boxes"]
            labels = outputs["pred_logits"]
            print(
                f"PyTorch inference output shapes - Boxes: {dets.shape}, Labels: {labels.shape}"
            )
    model.cpu()
    input_tensors = input_tensors.cpu()

    # Export to ONNX
    output_file = export_onnx(
        output_dir=output_dir,
        model=model,
        input_names=input_names,
        input_tensors=input_tensors,
        output_names=output_names,
        dynamic_axes=dynamic_axes,
        backbone_only=backbone_only,
        verbose=verbose,
        opset_version=opset_version,
    )

    print(f"Successfully exported ONNX model to: {output_file}")

    if simplify:
        sim_output_file = onnx_simplify(
            onnx_dir=output_file,
            input_names=input_names,
            input_tensors=input_tensors,
            force=force,
        )
        print(f"Successfully simplified ONNX model to: {sim_output_file}")

    print("ONNX export completed successfully")
    self.model = self.model.to(device)


def forward_export(self, tensors):
    srcs, _, poss = self.backbone(tensors)
    # only use one group in inference
    refpoint_embed_weight = self.refpoint_embed.weight[: self.num_queries]
    query_feat_weight = self.query_feat.weight[: self.num_queries]

    hs, ref_unsigmoid, hs_enc, ref_enc = self.transformer(
        srcs, None, poss, refpoint_embed_weight, query_feat_weight
    )

    if self.bbox_reparam:
        outputs_coord_delta = self.bbox_embed(hs)
        outputs_coord_cxcy = (
            outputs_coord_delta[..., :2] * ref_unsigmoid[..., 2:]
            + ref_unsigmoid[..., :2]
        )
        outputs_coord_wh = (
            outputs_coord_delta[..., 2:].exp() * ref_unsigmoid[..., 2:]
        )
        outputs_coord = torch.concat(
            [outputs_coord_cxcy, outputs_coord_wh], dim=-1
        )
    else:
        outputs_coord = (self.bbox_embed(hs) + ref_unsigmoid).sigmoid()
    outputs_class = self.class_embed(hs)
    prob = outputs_class.sigmoid()
    # Static topk with explicit self.num_queries
    topk_values, topk_indexes = torch.topk(
        prob.view(outputs_class.shape[0], -1), self.num_queries, dim=1
    )
    scores = topk_values
    b_size = prob.shape[0]
    # Fix type conversion - apply to the result, not the divisor
    topk_boxes = topk_indexes // outputs_class.shape[2]
    labels = (
        (topk_indexes % outputs_class.shape[2])
        .view(b_size, self.num_queries)
        .to(torch.int32)
    )
    boxes = box_ops.box_cxcywh_to_xyxy(outputs_coord)
    boxes = torch.gather(boxes, 1, topk_boxes.unsqueeze(-1).repeat(1, 1, 4))

    # Create scale factor with explicit reshape to ensure static shape
    img_h, img_w = tensors.shape[2].unsqueeze(dim=0).expand(
        1, self.num_queries
    ), tensors.shape[3].unsqueeze(dim=0).expand(1, self.num_queries)
    # Ensure static shapes with reshape operations
    scale_fct = torch.cat([img_w, img_h, img_w, img_h], dim=1).view(
        b_size, self.num_queries, 4
    )
    # Ensure boxes has static shape by forcing dimension
    boxes = boxes * scale_fct
    boxes = boxes.reshape(b_size, self.num_queries, 4)
    scores = scores.reshape(b_size, self.num_queries)

    # Create num_dets with static shape
    num_dets = torch.full(
        (b_size, 1), self.num_queries, dtype=torch.int32, device=boxes.device
    ).view((b_size, 1))

    # Return with explicit static shapes
    return num_dets, boxes, scores, labels


def interpolate_pos_encoding(
    self, embeddings: torch.Tensor, height: int, width: int
) -> torch.Tensor:
    """
    This method allows to interpolate the pre-trained position encodings, to be able to use the model on higher
    resolution images. This implementation supports torch.jit tracing while maintaining backwards compatibility
    with the original implementation.

    Adapted from:
    - https://github.com/facebookresearch/dino/blob/main/vision_transformer.py
    - https://github.com/facebookresearch/dinov2/blob/main/dinov2/models/vision_transformer.py
    """
    num_patches = embeddings.shape[1] - 1
    num_positions = self.position_embeddings.shape[1] - 1

    # Skip interpolation for matching dimensions (unless tracing)
    if (
        not torch.jit.is_tracing()
        and num_patches == num_positions
        and height == width
    ):
        return self.position_embeddings

    # Handle class token and patch embeddings separately
    class_pos_embed = self.position_embeddings[:, 0]
    patch_pos_embed = self.position_embeddings[:, 1:]
    dim = embeddings.shape[-1]

    # Calculate new dimensions
    height = height // self.config.patch_size
    width = width // self.config.patch_size

    # Reshape for interpolation
    sqrt_num_positions = torch_int(num_positions**0.5)
    patch_pos_embed = patch_pos_embed.reshape(
        1, sqrt_num_positions, sqrt_num_positions, dim
    )
    patch_pos_embed = patch_pos_embed.permute(0, 3, 1, 2)

    # Store original dtype for restoration after interpolation
    target_dtype = patch_pos_embed.dtype

    # Interpolate at float32 precision
    patch_pos_embed = nn.functional.interpolate(
        patch_pos_embed.to(dtype=torch.float32),
        size=(
            torch_int(height),
            torch_int(width),
        ),  # Explicit size instead of scale_factor
        mode="bicubic",
        align_corners=False,
        antialias=False,
    ).to(dtype=target_dtype)

    # Validate output dimensions if not tracing
    if not torch.jit.is_tracing():
        if (
            int(height) != patch_pos_embed.shape[-2]
            or int(width) != patch_pos_embed.shape[-1]
        ):
            raise ValueError(
                "Width or height does not match with the interpolated position embeddings"
            )

    # Reshape back to original format
    patch_pos_embed = patch_pos_embed.permute(0, 2, 3, 1).view(1, -1, dim)

    # Combine class and patch embeddings
    return torch.cat((class_pos_embed.unsqueeze(0), patch_pos_embed), dim=1)


def forward(self, srcs, masks, pos_embeds, refpoint_embed, query_feat):
    src_flatten = []
    mask_flatten = [] if masks is not None else None
    lvl_pos_embed_flatten = []
    spatial_shapes = torch.empty(
        (len(srcs), 2), device=srcs[0].device, dtype=torch.long
    )
    valid_ratios = [] if masks is not None else None
    for lvl, (src, pos_embed) in enumerate(zip(srcs, pos_embeds)):
        bs, c, h, w = src.shape
        spatial_shapes[lvl, 0] = h
        spatial_shapes[lvl, 1] = w

        src = src.flatten(2).transpose(1, 2)  # bs, hw, c
        pos_embed = pos_embed.flatten(2).transpose(1, 2)  # bs, hw, c
        lvl_pos_embed_flatten.append(pos_embed)
        src_flatten.append(src)
        if masks is not None:
            mask = masks[lvl].flatten(1)  # bs, hw
            mask_flatten.append(mask)
    memory = torch.cat(src_flatten, 1)  # bs, \sum{hxw}, c
    if masks is not None:
        mask_flatten = torch.cat(mask_flatten, 1)  # bs, \sum{hxw}
        valid_ratios = torch.stack([self.get_valid_ratio(m) for m in masks], 1)
    lvl_pos_embed_flatten = torch.cat(
        lvl_pos_embed_flatten, 1
    )  # bs, \sum{hxw}, c
    level_start_index = torch.cat(
        (spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1])
    )

    if self.two_stage:
        output_memory, output_proposals = gen_encoder_output_proposals(
            memory,
            mask_flatten,
            spatial_shapes,
            unsigmoid=not self.bbox_reparam,
        )
        # group detr for first stage
        refpoint_embed_ts, memory_ts, boxes_ts = [], [], []
        group_detr = self.group_detr if self.training else 1
        for g_idx in range(group_detr):
            output_memory_gidx = self.enc_output_norm[g_idx](
                self.enc_output[g_idx](output_memory)
            )

            enc_outputs_class_unselected_gidx = self.enc_out_class_embed[
                g_idx
            ](output_memory_gidx)
            if self.bbox_reparam:
                enc_outputs_coord_delta_gidx = self.enc_out_bbox_embed[g_idx](
                    output_memory_gidx
                )
                enc_outputs_coord_cxcy_gidx = (
                    enc_outputs_coord_delta_gidx[..., :2]
                    * output_proposals[..., 2:]
                    + output_proposals[..., :2]
                )
                enc_outputs_coord_wh_gidx = (
                    enc_outputs_coord_delta_gidx[..., 2:].exp()
                    * output_proposals[..., 2:]
                )
                enc_outputs_coord_unselected_gidx = torch.concat(
                    [enc_outputs_coord_cxcy_gidx, enc_outputs_coord_wh_gidx],
                    dim=-1,
                )
            else:
                enc_outputs_coord_unselected_gidx = (
                    self.enc_out_bbox_embed[g_idx](output_memory_gidx)
                    + output_proposals
                )  # (bs, \sum{hw}, 4) unsigmoid

            topk = min(
                self.num_queries, enc_outputs_class_unselected_gidx.shape[-2]
            )
            topk_proposals_gidx = torch.topk(
                enc_outputs_class_unselected_gidx.max(-1)[0], topk, dim=1
            )[
                1
            ]  # bs, nq

            refpoint_embed_gidx_undetach = torch.gather(
                enc_outputs_coord_unselected_gidx,
                1,
                topk_proposals_gidx.unsqueeze(-1).repeat(1, 1, 4),
            )  # unsigmoid
            # for decoder layer, detached as initial ones, (bs, nq, 4)
            refpoint_embed_gidx = refpoint_embed_gidx_undetach.detach()

            # get memory tgt
            tgt_undetach_gidx = torch.gather(
                output_memory_gidx,
                1,
                topk_proposals_gidx.unsqueeze(-1).repeat(1, 1, self.d_model),
            )

            refpoint_embed_ts.append(refpoint_embed_gidx)
            memory_ts.append(tgt_undetach_gidx)
            boxes_ts.append(refpoint_embed_gidx_undetach)
        # concat on dim=1, the nq dimension, (bs, nq, d) --> (bs, nq, d)
        refpoint_embed_ts = torch.cat(refpoint_embed_ts, dim=1)
        # (bs, nq, d)
        memory_ts = torch.cat(memory_ts, dim=1)  # .transpose(0, 1)
        boxes_ts = torch.cat(boxes_ts, dim=1)  # .transpose(0, 1)

    tgt = query_feat.unsqueeze(0).repeat(bs, 1, 1)
    refpoint_embed = refpoint_embed.unsqueeze(0).repeat(bs, 1, 1)
    if self.two_stage:
        ts_len = refpoint_embed_ts.shape[-2]
        refpoint_embed_ts_subset = refpoint_embed[..., :ts_len, :]
        refpoint_embed_subset = refpoint_embed[..., ts_len:, :]

        if self.bbox_reparam:
            refpoint_embed_cxcy = (
                refpoint_embed_ts_subset[..., :2] * refpoint_embed_ts[..., 2:]
            )
            refpoint_embed_cxcy = (
                refpoint_embed_cxcy + refpoint_embed_ts[..., :2]
            )
            refpoint_embed_wh = (
                refpoint_embed_ts_subset[..., 2:].exp()
                * refpoint_embed_ts[..., 2:]
            )
            refpoint_embed_ts_subset = torch.concat(
                [refpoint_embed_cxcy, refpoint_embed_wh], dim=-1
            )
        else:
            refpoint_embed_ts_subset = (
                refpoint_embed_ts_subset + refpoint_embed_ts
            )

        refpoint_embed = torch.concat(
            [refpoint_embed_ts_subset, refpoint_embed_subset], dim=-2
        )

    hs, references = self.decoder(
        tgt,
        memory,
        memory_key_padding_mask=mask_flatten,
        pos=lvl_pos_embed_flatten,
        refpoints_unsigmoid=refpoint_embed,
        level_start_index=level_start_index,
        spatial_shapes=spatial_shapes,
        valid_ratios=(
            valid_ratios.to(memory.dtype)
            if valid_ratios is not None
            else valid_ratios
        ),
    )
    if self.two_stage:
        if self.bbox_reparam:
            return hs, references, memory_ts, boxes_ts
        else:
            return hs, references, memory_ts, boxes_ts.sigmoid()
    return hs, references, None, None


def export_onnx_main(
    output_dir,
    model,
    input_names,
    input_tensors,
    output_names,
    dynamic_axes,
    backbone_only=False,
    verbose=True,
    opset_version=17,
):
    export_name = "model"
    output_file = os.path.join(output_dir, f"{export_name}.onnx")

    # Prepare model for export
    if hasattr(model, "export"):
        model.export()

    torch.onnx.export(
        model,
        input_tensors,
        output_file,
        input_names=input_names,
        output_names=output_names,
        export_params=True,
        keep_initializers_as_inputs=False,
        do_constant_folding=True,
        verbose=verbose,
        opset_version=opset_version,
        dynamic_axes=dynamic_axes,
    )

    print(f"\nSuccessfully exported ONNX model: {output_file}")
    return output_file


def onnx_simplify_main(onnx_dir: str, input_names, input_tensors, force=True):
    sim_onnx_dir = onnx_dir

    if isinstance(input_tensors, torch.Tensor):
        input_tensors = [input_tensors]

    print(f"start simplify ONNX model: {onnx_dir}")
    opt = OnnxOptimizer(onnx_dir)
    opt.info("Model: original")
    opt.common_opt()
    opt.info("Model: optimized")
    opt.save_onnx(sim_onnx_dir)
    input_dict = {
        name: tensor.detach().cpu().numpy()
        for name, tensor in zip(input_names, input_tensors)
    }
    model_opt, check_ok = onnxsim.simplify(
        onnx_dir, check_n=3, input_data=input_dict, dynamic_input_shape=False
    )
    if check_ok:
        onnx.save(model_opt, sim_onnx_dir)
    else:
        raise RuntimeError("Failed to simplify ONNX model.")
    print(f"Successfully simplified ONNX model: {sim_onnx_dir}")
    return sim_onnx_dir
